from .api import TransitionApiView
from .transition import DetailsListApiView , ActionListApi, WorkEventsListApi, WorkFlowitemsListApi
from django.urls import path



urlpatterns = [
    path('action/',ActionListApi.as_view()),
    path('model/',DetailsListApiView.as_view()),
    path('workflowitems/',WorkFlowitemsListApi.as_view()),
    path('workevents/',WorkEventsListApi.as_view()),
    path('api/transition/',TransitionApiView.as_view())
]



from asyncore import read
from rest_framework import serializers
from .models import TransitionManager , workevents , workflowitems , Action
from django.conf import settings
from django.apps import apps


class Workeventsserializer(serializers.ModelSerializer):
    event_user = serializers.SlugRelatedField(read_only=True, slug_field = 'email' or 'id')
    class Meta:
        model = workevents
        fields = [
            'id',
            'action',
            'subaction',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'from_party',
            'to_party'
        ]




class Workitemserializer(serializers.ModelSerializer):
    WorkFlowEvents = Workeventsserializer(many=True, read_only=True)
    event_user = serializers.SlugRelatedField(read_only=True, slug_field = 'email' or 'id')
    class Meta:
        model = workflowitems
        fields = [
            'id',
            'transitionmanager',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'current_from_party',
            'current_to_party',
            'WorkFlowEvents',
        ]




class TransitionManagerserializer(serializers.ModelSerializer):
    workflowitems = Workitemserializer(read_only = True)
    model = serializers.SerializerMethodField()
    wf_item_id = serializers.SerializerMethodField()
    class Meta:
        model = TransitionManager
        fields = [
            'id',
            'type',
            't_id',
            'wf_item_id',
            'model',
            'workflowitems'
        ]

    def get_model(self,obj):
        try:
            arr = settings.FINFLO['WORK_MODEL']
            for i in arr:
                user = apps.get_model(i)
                qs = user.objects.filter(id = obj.t_id ).values()
                if qs.exists():
                    break
            return qs
        except:
            return None
                    

    def get_wf_item_id(self,obj):
        return obj.workflowitems.id



class Actionseriaizer(serializers.ModelSerializer):
    class Meta:
        model = Action
        fields = '__all__'



class workflowitemslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workflowitems
        fields = '__all__'


class workeventslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workevents
        fields = '__all__'


from ast import excepthandler
from xml.dom import ValidationErr
from django.db import models
from .middleware import get_current_user
from .validators import validate_postive
from django.core.exceptions import ValidationError



try:
   from django.conf import settings
except ImportError:
    raise Exception("settings.py file is required to run this project")



## CORE CONFIGURATIONS CLASSES ##


class Temp(models.Model):
    pass


class Flowmodel(models.Model):
    description = models.CharField(max_length=255)

    def __str__(self):
        return self.description

    class Meta:
        verbose_name_plural = "1. FlowModel"


# Create your models here.

class TransitionManager(models.Model):
    t_id = models.IntegerField()
    type = models.CharField(max_length = 255 )
    sub_sign = models.IntegerField(default = 0 )

    # default 1 for initial submit and maker process

    def save(self, *args, **kwargs):
        self.type = self.type.upper()
        return super(TransitionManager,self).save(*args, **kwargs)
    
    def __str__(self):
        return "{0}_{1}".format(self.type.lower() , self.id)


class SignList(models.Model):
    sign_id = models.IntegerField()
    name = models.CharField(max_length = 255 , default = None)
    sub_name = models.CharField(max_length = 255 , default = None)

    def __str__(self):
        return self.name
    

class States(models.Model):
    description = models.CharField(max_length=255)
    
    class Meta:
        verbose_name_plural = "2. States"

    def __str__(self):
        return self.description


class Action(models.Model):
    description = models.CharField(max_length=255 , blank = True , null=True , help_text = 'e.g., SUBMIT , DELETE') 
    model = models.ForeignKey(Flowmodel , on_delete = models.CASCADE , blank = True , null = True )
    from_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , related_name = 'action_from_state' , blank = True , null = True , help_text = "initial from state for transition ")
    to_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , blank = True , null = True , help_text = "final state for the transition to take place")
    # optional fields
    intermediator = models.BooleanField(default = False)
    from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , help_text = 'this field is optional' , related_name = 'from_transition_party')
    to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , help_text = 'this field is optional' , related_name = 'to_last_party')
    stage_required = models.ForeignKey(SignList , on_delete= models.DO_NOTHING ,blank = True , null = True ,help_text = "IMPORTANT : if 0 means initial_transition ")
    
    def save(self, *args, **kwargs):
        self.description = self.description.upper()
        # try:
        #     sign_len = SignList.objects.get(name = self.stage_required )
        #     self.sign_required = sign_len.sign_id
        # except:
        #     pass
        return super(Action,self).save(*args, **kwargs)

    def __str__(self):
        return "{0} -> {1}".format(self.description , self.model)

    class Meta:
        verbose_name_plural = "3. Action"
        unique_together = ('description', 'model') 
 

class workflowitems(models.Model):
    
    created_date = models.DateTimeField(auto_now_add=True)
    transitionmanager = models.OneToOneField(TransitionManager, on_delete=models.CASCADE,blank=True, null=True )
    initial_state  = models.CharField(max_length=50,default = 'DRAFT')
    interim_state = models.CharField(max_length=50,default = 'DRAFT')
    final_state = models.CharField(max_length=50,default = 'DRAFT') 
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.DO_NOTHING , blank=True, null=True)
    next_available_transitions =  models.CharField(max_length=500)
    current_from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False , related_name = 'wf_from_party')
    current_to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False )
    action = models.CharField(max_length=25 , blank=True, null=True , default = 'DRAFT')
    subaction = models.CharField(max_length=55 , blank=True, null=True)
    previous_action = models.CharField(max_length=55 , blank=True, null=True)
    model_type  = models.CharField(max_length=55, blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    
    class Meta:
        verbose_name_plural = "WorkFlowItem"
        ordering = ['id']

    # def save(self, *args, **kwargs):
    #    if self.from_party is None:
    #         self.from_party = None
    #    super(workflowitems, self).save(*args, **kwargs) # Call the real save() method


# WORKEVENTS
class workevents(models.Model):

    workflowitems = models.ForeignKey(workflowitems, on_delete=models.CASCADE , related_name='WorkFlowEvents')
    action = models.CharField(max_length=25, blank=True, null=True , default = 'DRAFT')
    subaction = models.CharField(max_length=55 , blank=True, null=True , default = 'DRAFT')
    initial_state  = models.CharField(max_length=50 , default = 'DRAFT')
    interim_state = models.CharField(max_length=50,default = 'DRAFT')
    final_state = models.CharField(max_length=50,default = 'DRAFT')
    from_party = models.ForeignKey(settings.FINFLO['FROM_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False ,related_name = 'we_from_party')
    to_party = models.ForeignKey(settings.FINFLO['TO_PARTY'] or Temp , models.SET_NULL , blank = True , null = True , editable = False )
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL,  on_delete=models.DO_NOTHING , blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True)
    type = models.CharField(max_length=55, blank=True, null=True)
    

    class Meta:
        verbose_name_plural = "WorkFlowEvent"
        ordering = ['id']








# Generated by Django 3.2.5 on 2022-08-25 07:28
from enum import Enum
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
from finflo.validators import validate_postive



# DEFAULT ENUM CLASS FOR ACTION AND STATES 
class Values(Enum):
    NO_SIGN = "NO_SIGN"
    DRAFT = "DRAFT"
    RETURN = "RETURN"




# RUN ON MIGRATIONS ONLY

def create_flow_model(apps, schema_editor):
    flowmodel = apps.get_model("finflo.Flowmodel")
    try:
        a = settings.FINFLO['WORK_MODEL']
        db_alias = schema_editor.connection.alias
        for i in a:
            flowmodel.objects.using(db_alias).create(description = i)
    except:
        raise LookupError("Couldn't find work_model in settings.py check documentation")



def create_sign_list(apps, schema_editor):
    states = apps.get_model("finflo.SignList")
    db_alias = schema_editor.connection.alias
    states.objects.using(db_alias).create(name = Values.NO_SIGN, sign_id = 1 , sub_name = Values.NO_SIGN)


def create_initial_draft(apps, schema_editor):
    states = apps.get_model("finflo.States")
    db_alias = schema_editor.connection.alias
    states.objects.using(db_alias).create(description = Values.DRAFT)
    


def creates_default_return_Action(apps,schema_editor):
    action = apps.get_model('finflo.Action')
    states = apps.get_model("finflo.States")
    db_alias = schema_editor.connection.alias
    action.objects.using(db_alias).create(description = Values.RETURN , to_state = states.objects.get(id = 1 ) )





class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('accounts', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Flowmodel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': '1. FlowModel',
            },
        ),
        migrations.RunPython(create_flow_model),
        migrations.CreateModel(
            name='SignList',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sign_id', models.IntegerField()),
                ('name', models.CharField(default=None, max_length=255)),
                ('sub_name', models.CharField(default=None, max_length=255)),
            ],
        ),
        migrations.RunPython(create_sign_list),
        migrations.CreateModel(
            name='States',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': '2. States',
            },
        ),
        migrations.RunPython(create_initial_draft),
        migrations.CreateModel(
            name='Temp',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='TransitionManager',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('t_id', models.IntegerField()),
                ('type', models.CharField(max_length=255)),
                ('sub_sign', models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='workflowitems',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('initial_state', models.CharField(default='DRAFT', max_length=50)),
                ('interim_state', models.CharField(default='DRAFT', max_length=50)),
                ('final_state', models.CharField(default='DRAFT', max_length=50)),
                ('next_available_transitions', models.CharField(max_length=500)),
                ('action', models.CharField(blank=True, default='DRAFT', max_length=25, null=True)),
                ('subaction', models.CharField(blank=True, max_length=55, null=True)),
                ('previous_action', models.CharField(blank=True, max_length=55, null=True)),
                ('model_type', models.CharField(blank=True, max_length=55, null=True)),
                ('comments', models.CharField(blank=True, max_length=500, null=True)),
                ('is_read', models.BooleanField(blank=True, default=True, null=True)),
                ('final_value', models.BooleanField(blank=True, default=False, null=True)),
                ('current_from_party', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='wf_from_party', to='accounts.parties')),
                ('current_to_party', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='accounts.parties')),
                ('event_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
                ('transitionmanager', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='finflo.transitionmanager')),
            ],
            options={
                'verbose_name_plural': 'WorkFlowItem',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='workevents',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action', models.CharField(blank=True, default='DRAFT', max_length=25, null=True)),
                ('subaction', models.CharField(blank=True, default='DRAFT', max_length=55, null=True)),
                ('initial_state', models.CharField(default='DRAFT', max_length=50)),
                ('interim_state', models.CharField(default='DRAFT', max_length=50)),
                ('final_state', models.CharField(default='DRAFT', max_length=50)),
                ('is_read', models.BooleanField(blank=True, default=True, null=True)),
                ('final_value', models.BooleanField(blank=True, default=False, null=True)),
                ('comments', models.CharField(blank=True, max_length=500, null=True)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('type', models.CharField(blank=True, max_length=55, null=True)),
                ('event_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
                ('from_party', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='we_from_party', to='accounts.parties')),
                ('to_party', models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='accounts.parties')),
                ('workflowitems', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='WorkFlowEvents', to='finflo.workflowitems')),
            ],
            options={
                'verbose_name_plural': 'WorkFlowEvent',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, help_text='e.g., SUBMIT , DELETE', max_length=255, null=True)),
                ('intermediator', models.BooleanField(default=False)),
                ('sign_required', models.IntegerField(default=0, help_text='IMPORTANT : if 0 means initial_transition ', validators=[validate_postive])),
                ('from_party', models.ForeignKey(blank=True, help_text='this field is optional', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='from_transition_party', to='accounts.parties')),
                ('from_state', models.ForeignKey(blank=True, help_text='initial from state for transition ', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='action_from_state', to='finflo.states')),
                ('model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='finflo.flowmodel')),
                ('stage_required', models.ForeignKey(help_text='IMPORTANT : if 0 means initial_transition ',null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='finflo.signlist')),
                ('to_party', models.ForeignKey(blank=True, help_text='this field is optional', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='to_last_party', to='accounts.parties')),
                ('to_state', models.ForeignKey(blank=True, help_text='final state for the transition to take place', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='finflo.states')),
            ],
            options={
                'verbose_name_plural': '3. Action',
                'unique_together': {('description', 'model')},
            },
        ),
        migrations.RunPython(creates_default_return_Action),
    ]



from django.contrib import admin
from .models import Action, States, TransitionManager , Flowmodel , SignList
# Register your models here.
from django.conf import settings




admin.site.register(Flowmodel)
admin.site.register(TransitionManager)
admin.site.register(Action )
admin.site.register(States)
admin.site.register(SignList) 



from django import dispatch
from django.db.models.signals import post_save , pre_save , post_delete

from .middleware import get_current_user
from .models import TransitionManager, workevents, workflowitems , Action
from django.conf import settings




# CUSTOM MULTIPLE RECEIVER SIGNALS 


def finflo_receiver(signal, senders ,  **kwargs):
    def decorator(receiver_func):
        for sender in senders:
            if isinstance(signal, (list, tuple)):
                for s in signal:
                    s.connect(receiver_func, sender=sender, **kwargs)
            else:
                signal.connect(receiver_func, sender=sender,**kwargs)

        return receiver_func

    return decorator

# 



# MAIN SIGNAL RECEIVER FOR TRANSITION HANDLING - 4/8/2022 by anand


a = [ite for ite in settings.FINFLO['WORK_MODEL']]
@finflo_receiver(post_save, senders = a)
def create(sender, instance, **kwargs):
        a = str(sender)
        remove_characters = ["class", "models.","'","<",">"," "]
        for i in remove_characters:
            a = a.replace(i,"")
        manager = TransitionManager.objects.create(
            type = a.lower() , t_id = instance.id
        )
        action = Action.objects.get(model__description = a.lower())
        wf = workflowitems.objects.create(transitionmanager = manager , event_user = get_current_user() , initial_state = action.from_state.description , final_state = action.to_state.description)
        we = workevents.objects.create(workflowitems = wf , event_user = get_current_user() , initial_state = action.from_state.description , final_state = action.to_state.description)
        wf.save()
        we.save()
        manager.save()




from transition import FinFlotransition
from django.shortcuts import HttpResponse



# EXAMPLE CODE
qs = FinFlotransition()

def index(self):
    # transition2(stage = 3)
    qs.transition(type = "INVOICE" , action = "DELETE" , stage = 2)
    return HttpResponse(str("data"))



from django.http import HttpResponse
from .models import SignList, States, TransitionManager , Action, workevents, workflowitems , Flowmodel
from .serializer import TransitionManagerserializer , Actionseriaizer, Workitemserializer, workeventslistserializer, workflowitemslistserializer
from .middleware import get_current_user
from rest_framework.generics import ListAPIView , ListCreateAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework import status
from .exception import ModelNotFound, SignaturesNotMatching, TransitionNotAllowed
from .states import sign_list, sub_action
from django.db.models import Q
from rest_framework.exceptions import APIException
from django.conf import settings
from collections import deque

####################################################
#############       CORE     #######################
####################################################


## slice the interim state in wf_itemds







class FinFlotransition:


    def __init__(self, returns , next_value):
        self.returns = returns
        self.next_value = next_value
        super().__init__(returns , next_value)

    @property
    def returns_name(self):
        return self.returns.__name__


    def gets_default_return_action():
        qs = Action.objects.get(id = 1)
        return qs.description

    

    def gets_wf_item(gets_model):
        ws = workflowitems.objects.get(transitionmanager=gets_model.id)
        return ws

    # initiate a special function for reverse transition

    def return_transition():
        return None 

    # interim action for the transition handling
    def intermediator_transition_handler(self, type, action, t_id):
        try:
            gets_model = TransitionManager.objects.get(type=type.upper() , t_id = t_id)
            gets_flows = Flowmodel.objects.get(description  = type.lower())
            gets_action = Action.objects.get(description=action,model = gets_flows.id)
            gets_wf  = FinFlotransition.gets_wf_item(gets_model)
            sign_lists = []
            for item in SignList.objects.all():
                sign_lists.append(item.name)
                if item.name == gets_action.stage_required.name:
                    break
        except:
            raise ModelNotFound()

        
        # in-transition for return action  - can be customizable
        if action == FinFlotransition.gets_default_return_action().lower():


            wf = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_wf.initial_state, "interim_state" : gets_action.to_state.description ,
                "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
            workevents.objects.create(workflowitems=wf, event_user=get_current_user(),  initial_state=gets_wf.initial_state,final_value = True , 
                    interim_state = gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        


        # len action and sub_sign
        
        elif gets_model.sub_sign <= gets_action.sign_required:
            def Transition_Handler():
                    gets_sign = gets_action.sign_required
                    
                    
                    if len(sign_lists)-1 != gets_model.sub_sign:
                        ws = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" :  sign_lists[1 + gets_model.sub_sign], 
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                        we = workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state = sign_lists[1 + gets_model.sub_sign], final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        gets_model.sub_sign += 1
                        gets_model.save()

                    elif len(sign_lists)-1 == int(gets_model.sub_sign):
                        ws = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" : gets_action.to_state.description ,
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                        workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,final_value = True , 
                                                  interim_state = gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        gets_model.sub_sign += 1
                        gets_model.save()
            return Transition_Handler()   
            
        else:
            raise TransitionNotAllowed()
        

    
    # core transition handler 
    def transition(self, type, action, t_id):
        try:
            gets_model = TransitionManager.objects.get(type=type.upper() , t_id = t_id)
            gets_flows = Flowmodel.objects.get(description  = type.lower())
            gets_action = Action.objects.get(description=action,model = gets_flows.id)
            gets_wf  = FinFlotransition.gets_wf_item(gets_model)
            sign_lists = []
            for item in SignList.objects.all():
                sign_lists.append(item.name)
                if item.name == gets_action.stage_required.name:
                    break
        except:
            raise ModelNotFound()


        # in-transition for return action  - can be customizable
        if action == FinFlotransition.gets_default_return_action().lower():


            wf = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_wf.initial_state, "interim_state" : gets_action.to_state.description ,
                "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
            workevents.objects.create(workflowitems=wf, event_user=get_current_user(),  initial_state=gets_wf.initial_state,final_value = True , 
                    interim_state = gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        


        # len action and sub_sign
        
        elif gets_model.sub_sign <= gets_action.sign_required:
            def Transition_Handler():
                    gets_sign = gets_action.sign_required
                    
                    
                    if len(sign_lists)-1 != gets_model.sub_sign:
                        ws = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" :  sign_lists[1 + gets_model.sub_sign], 
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                        we = workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,
                                                  interim_state = sign_lists[1 + gets_model.sub_sign], final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        gets_model.sub_sign += 1
                        gets_model.save()

                    elif len(sign_lists)-1 == int(gets_model.sub_sign):
                        ws = workflowitems.objects.update_or_create( transitionmanager=gets_model or gets_model.id, defaults= {"initial_state" : gets_action.from_state.description, "interim_state" : gets_action.to_state.description ,
                            "final_state" : gets_action.to_state.description, "action" : action, "subaction" : sign_list[1 + gets_model.sub_sign], "model_type" : type.upper(), "event_user" : get_current_user() , "current_from_party" : gets_action.from_party , "current_to_party" : gets_action.to_party})
                        workevents.objects.create(workflowitems=gets_wf, event_user=get_current_user(),  initial_state=gets_action.from_state.description,final_value = True , 
                                                  interim_state = gets_action.to_state.description, final_state=gets_action.to_state.description, action=action, subaction=sub_action[0], type=type.upper(), from_party = gets_action.from_party , to_party = gets_action.to_party)
                        gets_model.sub_sign += 1
                        gets_model.save()
            return Transition_Handler()   
            
        else:
            raise TransitionNotAllowed()






####################################################
################      API       ####################
####################################################



#  ALL WORK_MODEL TRANSITION LIST 


class DetailsListApiView(ListAPIView):
    queryset = TransitionManager.objects.all()
    serializer_class = TransitionManagerserializer
    permission_classes = [IsAuthenticated]
    
    def get_queryset(self):
        type = self.request.query_params.get('type',None)
        t_id = self.request.query_params.get('t_id',None)
        if t_id and type is not None :
            queryset = TransitionManager.objects.filter(type__icontains = type , t_id = int(t_id))
        elif type is not None and t_id is None:
            queryset = TransitionManager.objects.filter(type__icontains = type)
        else:
            queryset = TransitionManager.objects.all()
        return queryset

    def list(self, request):
        queryset = self.get_queryset()
        serializer = TransitionManagerserializer(queryset, many=True)
        return Response({"status": "success", "type" : settings.FINFLO['WORK_MODEL'] , "data": serializer.data}, status=status.HTTP_200_OK)



# WORFLOW API 

class WorkFlowitemsListApi(ListAPIView):
    queryset = workflowitems.objects.all()
    serializer_class = Workitemserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workflowitems.objects.all()
        serializer = Workitemserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)




# WORKEVENTS API 


class WorkEventsListApi(ListAPIView):
    queryset = workevents.objects.all()
    serializer_class = workeventslistserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = workevents.objects.all()
        serializer = workeventslistserializer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)



# ACTION CREATE AND LIST API 


class ActionListApi(ListCreateAPIView):
    queryset = Action.objects.all()
    serializer_class = Actionseriaizer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = Action.objects.all()
        serializer = Actionseriaizer(queryset, many=True)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)


    def post(self, request):
        serializer = Actionseriaizer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)









## BASIC EXCEPTION CLASSES 

from rest_framework.exceptions import APIException

class ModelNotFound(APIException):
    status_code = 204
    default_detail = 'Matching action and type not found , try with some other type '
    default_code = 'model_not_found'



class TransitionNotAllowed(APIException):
    status_code = 204
    default_detail = 'Transition Not Allowed .'
    default_code = 'transition_unavailable'



class SignaturesNotMatching(APIException):
    status_code = 204
    default_detail = 'Signatures not matching , only positive index accepted'
    default_code = 'signatures_index_not_matching'



class SignLengthError(APIException):
    status_code = 204
    default_detail = 'Signatures not matching , only positive index accepted'
    default_code = 'signatures_index_not_matching'


from rest_framework.views import APIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .transition import FinFlotransition

class TransitionApiView(APIView):
    permission_classes = [IsAuthenticated]

    def post(self,request):
        transitions = FinFlotransition()
        type = request.data.get("type")
        action = request.data.get("action")
        stage = request.data.get("stage")
        t_id = request.data.get("t_id")
        cs = int(stage) 
        if type and action  is not None:
            transitions.transition(type = type, action = action , stage = cs , id = t_id)
            return Response({"status" : "Transition success"},status = status.HTTP_200_OK)
        return Response({"status" : "Transition failure"},status = status.HTTP_204_NO_CONTENT)

